
//		Elevated.cpp - Implements an @ELEVATED function.


#define UNICODE					1
#define _UNICODE				1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR				1
#define VER_MINOR				0
#define VER_BUILD				1
#define VER_PATCH				1

#define PLUGIN_NAME				L"Elevated"

#ifdef _WIN64
#define PLATFORM				L"x64"
#else
#define PLATFORM				L"x86"
#endif

#define MAX_ARG_LEN				8192
#define SHORT_BUF_LEN			1024
#define MAX_FILENAME_LEN		32768

#define ERRHANDLE				GetStdHandle( STD_ERROR_HANDLE )


HINSTANCE MyDLLHandle			= NULL;

OSVERSIONINFOEXW VerInfo;



const wchar_t PluginFeatures[] = L"@elevated";



const wchar_t F_ElevatedHelpText[]	= L"@ELEVATED[pid] \r\n\
\n\
Tests whether a process is elevated:\n\
\n\
\t 1\televated\n\
\t 0\tnot elevated\n\
\t-1\terror\n\
\n\
The pid may be entered in decimal, or hexadecimal\n\
with a leading 0x.  If pid is not specified, the current \n\
instance of TCC is assumed.\n";




const int NumHelpTexts							= 1;
const wchar_t HelpFeature[NumHelpTexts][11]		= { L"@elevated" };
const wchar_t *HelpText[NumHelpTexts]			= { F_ElevatedHelpText };



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



BOOL GetVersion2( OSVERSIONINFOEXW *VerInfo )
{
	if ( !VerInfo )
		return -1;


	typedef NTSTATUS ( WINAPI _RtlGetVersionProc ) ( OSVERSIONINFOEXW *VerInfo );
	HMODULE NtDllHandle						= GetModuleHandle( L"NtDll.dll" );
	_RtlGetVersionProc *RtlGetVersionProc	= NULL;


	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	if ( NtDllHandle )
		RtlGetVersionProc = (_RtlGetVersionProc*) GetProcAddress( NtDllHandle, "RtlGetVersion" );

	if ( RtlGetVersionProc ) {
		RtlGetVersionProc( VerInfo );
		if ( VerInfo->dwMajorVersion >= 6 )
			return TRUE;
	}

	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	return ( GetVersionEx( (OSVERSIONINFOW*) VerInfo ) != 0 );
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}


DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/elevated.html";
	piInfo.pszDescription	= (LPTSTR) L"Elevated plugin";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}


DLLExports BOOL WINAPI InitializePlugin( void )
{
	GetVersion2( &VerInfo );

	if ( VerInfo.dwMajorVersion < 6 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Windows Vista or later is required.  Plugin not loaded.\r\n" );
		return -1;
	}


	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( wcscmp( shortbuf, L"00" ) == 0 )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}


DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{

#ifdef NEWHELP_VERSION
	CloseHelpWindow();
#endif

	return 0;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( ( EndChar == NULL ) && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar != NULL )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( ( EndChar == NULL ) && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar != NULL )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



//	Strips off any leading or trailing spaces from a string.

int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



int ProcessElevated( DWORD Pid )
{
	HANDLE ProcessHandle		= NULL;
	HANDLE ProcessToken			= NULL;
	TOKEN_ELEVATION Elevated	= { 0 };
	DWORD ReturnSize			= 0;

	int rv						= -1;


	ProcessHandle = OpenProcess( PROCESS_QUERY_LIMITED_INFORMATION, FALSE, Pid );
	if ( ProcessHandle ) {
		//	wprintf( L"   * ProcessHandle = 0x%08X\n", ProcessHandle );

		if ( OpenProcessToken( ProcessHandle, TOKEN_QUERY, &ProcessToken ) ) {
			//	wprintf( L"   * ProcessToken = 0x%08X\n", ProcessToken );

			if ( GetTokenInformation( ProcessToken, TokenElevation, &Elevated, sizeof( TOKEN_ELEVATION ), &ReturnSize ) )
				rv = Elevated.TokenIsElevated ? 1 : 0;

			CloseHandle( ProcessToken );
		}

		CloseHandle( ProcessHandle );
	}

	return rv;
}



DLLExports int WINAPI f_elevated( LPTSTR String )
{
	if ( !String )
		return 1;


	DWORD Pid				= 0;
	int rv					= 0;


	TrimLine( String );
	_wcsupr_s( String, MAX_ARG_LEN );


	if ( String[0] == '\0' )
		Pid = GetCurrentProcessId();
	else if ( ParseInt( String, (unsigned int*) &Pid, NULL ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad PID: \"%s\"\r\n", String );
		return -1;
	}


	rv = ProcessElevated( Pid );

	/*
	if ( rv < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening process: 0x%04X\r\n", Pid );
		return -1;
	}
	*/

	wsprintf( String, L"%d", ProcessElevated( Pid ) );
	return 0;
}




#ifndef NEWHELP_VERSION

//		Simpler MessageBox() help, useful for smaller plugins which don't merit a full HTML help system:

DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}

#endif		//	NEWHELP_VERSION
